// This file is part of TTDXRecolor.
// Copyright (c) 2010, Uwe Domaratius

#include "CColorOption.hpp"
#include "CPalette.hpp"
#include "COptions.hpp"

#include <boost/regex.hpp>
#include <boost/tokenizer.hpp>
#include <iostream>
#include <sstream>
#include <algorithm>

CColorOption::CColorOption()
: mName("")
, mIndices()
, mValues()
, mDosPalette(false)
{
}

const UInt32 CColorOption::isNameValid(COptions* options) const
{
   UInt32 error = 0;
   // a color must have a name
   if (mName == "")
   {
      error += E_COLOR_NO_NAME_DEFINED;
      printErrorMessage(E_COLOR_NO_NAME_DEFINED);
   }
 
   // name must not contain whitespaces
   boost::char_separator<char> separator(" ");
   typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
   tokenizer tokens(mName, separator);   
   UInt32 number = 0;
   for (tokenizer::iterator i = tokens.begin();
              i != tokens.end(); ++i)
   {
      number++;
   }
   if (number > 1)
   {
      error += E_COLOR_NAME_CONTAINS_WHITESPACES;
      printErrorMessage(E_COLOR_NAME_CONTAINS_WHITESPACES);
   }

   // there must be no other color with that name
   if (options->getColorOption(mName))
   {
      error += E_COLOR_NAME_IN_USE;
      printErrorMessage(E_COLOR_NAME_IN_USE);
   }
   return error;
}

const UInt32 CColorOption::isIndexValid(COptions*) const
{
   UInt32 error = 0;
   // at least one index or value must be specified
   if (mIndices.empty() && mValues.empty())
   {
      error += E_COLOR_NOTHING_DEFINED;
      printErrorMessage(E_COLOR_NOTHING_DEFINED);
   }

   // if color indices are given, check if they are in range
   if (!mIndices.empty())
   {
      CPalette palette(mDosPalette);
      for (UInt32 i = 0; i < mIndices.size(); i++)
      {
         if (mIndices[i] > 255)
         {
            error += E_COLOR_INDEX_OUT_OF_RANGE;
            printErrorMessage(E_COLOR_INDEX_OUT_OF_RANGE);
         }
         if ((error != 0) && (palette.getColor(mIndices[i]).isSpecial()))
         {
            std::cout << "Warning: color index " << mIndices[i]
                      << " is a special color. This might yield unexpected results!"
                      << std::endl;
         }
      }
   }
   return error;
}

const UInt32 CColorOption::isValueValid(COptions*) const
{
   UInt32 error = 0;
   if (!mValues.empty())
   {
      if (mValues.size() % 3 != 0)
      {
         error += E_COLOR_INCORRECT_VALUE_COUNT;
         printErrorMessage(E_COLOR_INCORRECT_VALUE_COUNT);
      }
      for (UInt32 i = 0; i < mValues.size(); i++)
      {
         if (mValues[i] > 255)
         {
            error += E_COLOR_VALUE_OUT_OF_RANGE;
            printErrorMessage(E_COLOR_VALUE_OUT_OF_RANGE);
         }
      }
      CPalette palette(mDosPalette);
      for (UInt32 i = 0; i < mValues.size(); i+=3)
      {
         const UInt32 index = palette.getIndex(mValues[i],mValues[i+1],mValues[i+2]);
         if (index == INT_MAX)
         {
            error += E_COLOR_NOT_IN_PALETTE;
            printErrorMessage(E_COLOR_NOT_IN_PALETTE);
         }
         if (index < 256 && palette.getColor(index).isSpecial())
         {
            std::cout << "Warning: color (" << mValues[i] << "," 
                                            << mValues[i+1] << "," 
                                            << mValues[i+2] << ") is a special color. This might yield unexpected results!" 
                                            << std::endl;
         }
      }
   } 
   return error;
}

bool CColorOption::isValid(COptions* options) const
{
   // check the name for validity
   UInt32 errors = isNameValid(options);
   
   // check the given indices
   errors += isIndexValid(options);
   // check the given values
   errors += isValueValid(options);
    
   return 0 == errors;
}

void CColorOption::finalize(COptions* /* options */)
{
   // add all information from the value vector to the index vector, the value vector
   // is not needed for further processing
   CPalette palette(mDosPalette);
   for (UInt32 i = 0; i < mValues.size(); i+= 3)
   {
      const UInt32 index = palette.getIndex(mValues[i],mValues[i+1],mValues[i+2]);
      // check if that index is already given, then we just ignore it
      const std::vector<UInt32>::iterator result = find(mIndices.begin(),mIndices.end(),index);
      if (result == mIndices.end())
      {
         mIndices.push_back(index);
      }
   }
   // now simply clear the value vector, it is not needed anymore
   mValues.clear();
}

void CColorOption::addColorDefinition(const std::string& type, const std::string& value)
{
   // tokenize using whitespaces
   boost::char_separator<char> separator(" ");
   typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
   tokenizer tokens(value, separator);

   for (tokenizer::iterator i = tokens.begin();
        i != tokens.end(); ++i)
   {
      if (type == "index")
      {
         mIndices.push_back(NConvert::string2Int(*i));
      }
      else
      {
         mValues.push_back(NConvert::string2Int(*i));
      }
   }
}

void CColorOption::addLine(std::string& line)
{
   
   // a line always has to have the format option = value
   // non-matching lines are ignored
   const boost::regex option("([a-zA-Z]+)( ?)=( ?)(.+)");
   boost::smatch match;

   if (boost::regex_match(line, match, option))
   {
      std::string t = match[1].str();
      // I don't know why, but lint complains that it does not know std::transform
      static_cast<void>(std::transform(t.begin(), t.end(), t.begin(), tolower));//lint !e1025
      std::string t2 = match[4].str();
      static_cast<void>(std::transform(t2.begin(), t2.end(), t2.begin(), tolower));//lint !e1025
      if (t == "name")
      {
         mName = match[4].str();
      }
      if (t == "index" || t == "value")
      {
         addColorDefinition(t,t2);
      }
      if (t == "palette")
      {
         if (t2 == "dos")
         {
            mDosPalette = true;
         }
         if (t2 == "win")
         {
            mDosPalette = false;
         }
      }
   }
}
