// This file is part of TTDXRecolor.
// Copyright (c) 2010, Uwe Domaratius

#include "CImage.hpp"
#include "GTypes.hpp"
#include "CPalette.hpp"
#include <iostream>
#include <fstream>
#include <vector>
#include <boost/regex.hpp>

CImage::CImage(const std::string& filename)
: mFileName(filename)
, mIsValid(true)
, mDosPalette(false)
, mHeader()
, mOriginal()
{
   init();
}

void CImage::init(void)
{
   boost::regex pcx(".pcx$");
   if (!boost::regex_search(mFileName,pcx))
   {
      std::cout << "Error: filename does not end with .pcx" << std::endl;
      mIsValid = false;
   }
   else
   {
      std::ifstream ifs(mFileName.c_str(),std::ifstream::binary);
      if (!ifs.is_open())
      {
         std::cout << "Error: file " << mFileName << " could not be opened." << std::endl;
         mIsValid = false;
      }
      else
      {
         checkFileConsistency(ifs);
      }
      ifs.close();
   }
}

void CImage::checkHeader(const pcxHeader* const header)
{
   if (header)
   {
      // try to read the header
      if (0x0A != header->magicNumber)
      {
         std::cout << "Error: not a pcx file." << std::endl;
         mIsValid = false;
      }
      else
      if (5 != header->version)
      {
         std::cout << "Error: unsupported version." << std::endl;
         mIsValid = false;
      }
      else
      if (8 != header->bitsPerPixel)
      {
         std::cout << "Error: invalid color depth." << std::endl;
         mIsValid = false;
      }
      else
      if (1 != header->colorPlanes)
      {
         std::cout << "Error: unsupported number of color planes." << std::endl;
         mIsValid = false;
      }
   }
}

void CImage::storeOriginal(const char* const buffer, const Int32 length)
{
   if (mIsValid && buffer)
   {
      // create the header as member of the class
      mHeader.resize(128);
      for (UInt32 i = 0; i < 128; i++)
      {
         mHeader[i] = static_cast<UInt8>(buffer[i]);
      }
      mOriginal.resize(static_cast<UInt32>(length)-(128+769));
      for (UInt32 i = 0; i < static_cast<UInt32>(length-(128+769)); i++)
      {
         mOriginal[i] = static_cast<UInt8>(buffer[128+i]);
      }
   }
}

void CImage::checkFileConsistency(std::ifstream& ifs)
{
   char* buffer = 0;
   const Int32 length = readInputfile(buffer,ifs);
   if (length)
   {
      // this truncates the buffer, but that is ok here
      const pcxHeader* const header = reinterpret_cast<pcxHeader*>(buffer); //lint !e826 !e927
      checkHeader(header);
   }
   else
   {
      std::cout << "Error: Could not read from input file." << std::endl;
      mIsValid = false;
   }
   
   if (mIsValid)
   {
      // try to read the body
      // try to read the palette
      const Int32 offset = length-769;
      if (0x0C != buffer[static_cast<UInt32>(offset)])
      {
         std::cout << "Error: palette byte not found." << std::endl;
         mIsValid = false;
      }
      else
      {
         // this reinterpretation truncates buffer, but this is intended
         const pcxPalette* const palette = reinterpret_cast<pcxPalette*>(buffer+offset+1); //lint !e826 !e927
         if (!checkPalette(palette))
         {
            std::cout << "Error: palette mismatch, neither dos nor win." << std::endl;
            mIsValid = false;
         }
      }  
   }
   storeOriginal(buffer,length);
   // we're finished, return the result
   delete[] buffer;
}

bool CImage::checkPalette(const pcxPalette* const palette)
{
   if (palette)
   {
      CPalette winPalette(false);
      CPalette dosPalette(true);
      
      bool dos = true;
      bool win = true;
      for (UInt32 i = 0; i < 256; i++)
      {
         CColor c(palette->entry[3*i],palette->entry[3*i+1],palette->entry[3*i+2]);
         if (winPalette.getIndex(c) != i)
         {
            win = false;
         }
         if (dosPalette.getIndex(c) != i)
         {
            dos = false;
         }
      }
      // set flag for dos palette if necessary
      // not setting this flag with mIsValid being set denotes winPalette
      if (dos)
      {
         mDosPalette = true;
      }
      // positive result if its either dos or win
      return dos || win;
   }
   return false; // if palette == 0
}

const Int32 CImage::readInputfile(char*& buffer, std::ifstream& ifs) const
{
   if (ifs.is_open())
   {
      // get length of file:
      static_cast<void>(ifs.seekg (0, std::ios::end));
      const Int32 length = ifs.tellg();
      static_cast<void>(ifs.seekg (0, std::ios::beg));

      // allocate memory:
      buffer = new char [static_cast<UInt32>(length)];

      // read data as a block:
      static_cast<void>(ifs.read(buffer,length));
      ifs.close();
      return length;
   }
   return 0;
}

void CImage::encodeByte(std::vector<UInt8>& encoded, const UInt8 count, const UInt8 value) const
{
   if ((count > 1)||((value & 0xC0) == 0xC0))
   {
      encoded.push_back(static_cast<UInt8>(0xC0)+count);
      encoded.push_back(value);
   }
   else
   {
      encoded.push_back(value);
   }
}

void CImage::writeData(std::ofstream& ofs, const CRecolorTable& table) const
{
   CPalette from(mDosPalette);
   CPalette to(table.isDosPalette());
   std::vector<UInt8> encoded;
   UInt32 i = 0;
   while(i < mOriginal.size())
   {
      UInt8 byte = mOriginal[i];
      UInt8 count = 1;
      if (0xC0 == (byte & 0xC0))
      {
         // two topmost bytes set, this is a run-byte giving length
         count = byte & 0x3F;
         i++;
         byte = mOriginal[i];
      }
      const CColor c = from.getColor(byte);
      const UInt8 value = static_cast<UInt8>(table.getEntry(to.getIndex(c)));
      encodeByte(encoded,count,value);
      i++;
   }
   for (UInt32 j = 0; j < encoded.size(); j++)
   {
      ofs << std::hex << encoded[j];
   }
}

void CImage::writePalette(std::ofstream& ofs, const bool isDos) const
{
   CPalette palette(isDos);
   ofs << std::hex << static_cast<UInt8>(0x0C);
   for (UInt32 i = 0; i < 256; i++)
   {
      const CColor c = palette.getColor(i);
      ofs << std::hex << static_cast<UInt8>(c.r());
      ofs << std::hex << static_cast<UInt8>(c.g());
      ofs << std::hex << static_cast<UInt8>(c.b());
   }
}

void CImage::writeFile(const UInt32 number, const CRecolorTable& table)
{
   // take off the .pcx at the end and create new filename for output
   boost::regex pcx("(.*).pcx$");
   boost::smatch match;
   static_cast<void>(boost::regex_match(mFileName, match, pcx));
   std::string output(match[1].str());
   output += std::string("_");
   std::stringstream ss;
   std::string n;
   ss << number;
   ss >> n;
   output += n;
   output += std::string(".pcx");
   std::ofstream ofs(output.c_str(),std::ofstream::binary);
   if (ofs.is_open())
   {
      // write header
      for (UInt32 i = 0; i < mHeader.size(); i++)
      {
         ofs << std::hex << mHeader[i];
      }
      // encode and write output
      writeData(ofs,table);
      // write palette
      writePalette(ofs,table.isDosPalette());
      ofs.close();
   }
   else
   {
      std::cout << "Error: could write image file." << std::endl;
   }
}

void CImage::applyRecolor(const UInt32 number, const CRecolorTable &table)
{
   writeFile(number,table);
}
