// This file is part of TTDXRecolor.
// Copyright (c) 2010, Uwe Domaratius

#ifndef CIMAGE_HPP
#define CIMAGE_HPP

#include <string>
#include "GTypes.hpp"
#include "CRecolorTable.hpp"

//lint -esym(1712,CImage) no default constructor
class CImage
{
public:
   /** constructor
    *  @param filename name of the file to be used
    */
   explicit CImage(const std::string& fileName);

   /** get the validity state of the file
    *  @return true if file is valid image file, false otherwise
    */
   bool isValid(void) const;

   /** apply a recolor scheme
    *  @param index the number of the table to distinguish several tables based on the same input file
    *  @param table the recolor table to be applied
    */
   void applyRecolor(const UInt32 index, const CRecolorTable& table);
private:
   /// header of a pcx file, for easier work with the raw data
   struct pcxHeader
   {
      UInt8 magicNumber;
      UInt8 version;
      UInt8 compression;
      UInt8 bitsPerPixel;
      UInt16 coordinates[4];
      UInt16 DPIhorizontal;
      UInt16 DPIvertical;
      UInt8 colorMap[48];
      UInt8 reserved;
      UInt8 colorPlanes;
      UInt16 bytesPerRow;
      UInt16 paletteInfo;
      UInt16 imageWidth;
      UInt16 imageHeight;
      UInt8 empty[54];
   };

   struct pcxPalette
   {
      UInt8 entry[768];
   };

   /// the file name
   std::string mFileName;
   /// validity flag
   bool mIsValid;
   /// palette state (false for windows)
   bool mDosPalette;
   /// header of the file
   std::vector<UInt8> mHeader;
   /// buffer for the original image data
   std::vector<UInt8> mOriginal;
     
   /// initialize the file
   void init(void);
   /// check if the file is a valid pcx file
   void checkFileConsistency(std::ifstream& ifs);
   /// check if the header is from valid pcx
   void checkHeader(const pcxHeader* const header);
   /// check if the palette is a valid ttdx one
   bool checkPalette(const pcxPalette* const palette);
   /// read from input ifs into buffer, returns file size
   const Int32 readInputfile(char*& buffer, std::ifstream& ifs) const;
   /// write the changed data as file
   void writeFile(const UInt32 number, const CRecolorTable& table);
   /// encode the RLE scheme of pcx again
   void writeData(std::ofstream& ofs, const CRecolorTable& table) const;
   /// encode byte of RLE scheme
   void encodeByte(std::vector<UInt8>& encoded, const UInt8 count, const UInt8 value) const;
   /// write the palette
   void writePalette(std::ofstream& ofs, const bool isDos) const;
   /// store the original buffer data
   void storeOriginal(const char* const buffer, const Int32 length);
};

inline
bool CImage::isValid(void) const
{
   return mIsValid;
}

#endif
