// This file is part of TTDXRecolor.
// Copyright (c) 2010, Uwe Domaratius

#include "CMergeOption.hpp"
#include "CReplaceOption.hpp"
#include "COptions.hpp"
#include "GTypes.hpp"

#include <boost/tokenizer.hpp>
#include <boost/regex.hpp>

bool CMergeOption::countUsedIndices(const std::vector<std::string>& indices, std::vector<UInt32>& used, COptions* options) const
{
   for (UInt32 i = 0; i < indices.size(); i++)
   {
      const CReplaceOption* const r = options->getReplaceOption(indices[i]);
      if (r)
      {
         for (UInt32 j = 0; j < r->fromSize(); j++)
         {
            std::string hex("0x");
            hex += r->from(j);
            used[NConvert::string2Int(hex)]++;
         }
      }
   }
   for (UInt32 i = 0; i < used.size(); i++)
   {
      if (used[i] > 1)
      {
         return false;
      }
   }
   return true;
}

bool CMergeOption::checkGivenOptions(const std::vector<std::string>& indices,COptions* options) const
{
   bool result = true;
   for (UInt32 i = 0; i < indices.size(); i++)
   {
      if (!options->getReplaceOption(indices[i]))
      {
         result = false;
         break;
      }
   }
   return result;
}

bool CMergeOption::samePalette(const std::vector<std::string>& merge, COptions* options) const
{
   if (!merge.empty())
   {
      bool dos = false;
      bool win = false;
      for (UInt32 i = 0; i < merge.size(); i++)
      {
         const CReplaceOption* const c = options->getReplaceOption(merge[i]);
         if (c)
         {
            if (c->isDos())
            {
               dos = true;
            }
            else
            {
               win = true;
            }
         }
      }
      return ((win && !dos)||(!win && dos));
   }
   return true;
}

bool CMergeOption::isValid(COptions* options) const
{
   UInt32 error = 0;
   // check if at least two options have been given
   if (mMergeOptions.size() < 2 && mCombineOptions.size() < 2)
   {
      error += E_MERGE_OPTION_NUMBER;
      printErrorMessage(E_MERGE_OPTION_NUMBER);
   }
   // check if the options to be merged are actually known
   if (!checkGivenOptions(mMergeOptions,options) || !checkGivenOptions(mCombineOptions,options))
   {
      error += E_MERGE_OPTION_UNDEFINED;
      printErrorMessage(E_MERGE_OPTION_UNDEFINED);
   }

   // check if options can be merged or if they work on the same indices
   std::vector<UInt32> used;
   used.resize(256,0);
   if (!countUsedIndices(mMergeOptions,used,options))
   {
      error += E_MERGE_INDICES;
      printErrorMessage(E_MERGE_INDICES);
   }
   used.resize(0);
   used.resize(256,0);
   if (!countUsedIndices(mCombineOptions,used,options))
   {
      error += E_MERGE_INDICES;
      printErrorMessage(E_MERGE_INDICES);
   }

   if (!samePalette(mMergeOptions,options) || !samePalette(mCombineOptions,options))
   {
      error += E_MERGE_PALETTE;
      printErrorMessage(E_MERGE_PALETTE);
   }
   return 0 == error;
}

void CMergeOption::finalize(COptions* options)
{
   if (mCombineOptions.size() > 0)
   {
      CReplaceOption* const result = options->getReplaceOption(mCombineOptions[0]);
      for (UInt32 i = 1; i < mCombineOptions.size(); i++)
      {
         result->combine(options->getReplaceOption(mCombineOptions[i]));
      }
      for (UInt32 i = 1; i < mCombineOptions.size(); i++)
      {
         CReplaceOption* r = options->getReplaceOption(mCombineOptions[i]);
         options->erase(r);
         delete r;
      }
   }
   if (mMergeOptions.size() > 0)
   {
      CReplaceOption* const result = options->getReplaceOption(mMergeOptions[0]);
      for (UInt32 i = 1; i < mMergeOptions.size(); i++)
      {
         result->merge(options->getReplaceOption(mMergeOptions[i]));
      }
      for (UInt32 i = 1; i < mMergeOptions.size(); i++)
      {
         CReplaceOption* r = options->getReplaceOption(mMergeOptions[i]);
         options->erase(r);
         delete r;
      }
   }   
}

void CMergeOption::addLine(std::string& line)
{
   // a line always has to have the format option = value
   // non-matching lines are ignored
   const boost::regex option("([a-zA-Z]+)( ?)=( ?)(.+)");
   boost::smatch match;

   if (boost::regex_match(line, match, option))
   {
      std::string t = match[1].str();
      // I don't know why, but lint complains that it does not know std::transform
      static_cast<void>(std::transform(t.begin(), t.end(), t.begin(), tolower));//lint !e1025
      if (t == "merge" || t == "combine")
      {
         // tokenize match[2] using whitespaces
         const std::string value = match[4].str();
         boost::char_separator<char> separator(" ");
         typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
         tokenizer tokens(value, separator);

         for (tokenizer::iterator i = tokens.begin();
              i != tokens.end(); ++i)
         {
            if (t == "merge")
            {
               mMergeOptions.push_back(*i);
            }
            else
            {
               mCombineOptions.push_back(*i);
            }
         }
      }
   }
}
