// This file is part of TTDXRecolor.
// Copyright (c) 2010, Uwe Domaratius

#include "COptions.hpp"
#include "CColorOption.hpp"
#include "CReplaceOption.hpp"
#include "CMergeOption.hpp"
#include "CGenericOption.hpp"
#include "CImage.hpp"

// this pragma changes warning level to 3 for the header file, because it does not compile
// cleanly at warning level 4
#ifdef _MSC_VER
#pragma warning ( push , 3 )
#endif
#include <boost/program_options.hpp>
#ifdef _MSC_VER
#pragma warning ( pop )
#endif

#include <boost/regex.hpp>
#include <iostream>
#include <fstream>

COptions::COptions(int argc, char* argv[])
: mIsHelpRequested(false)
, mIsValid(true)
, mImageNames()
, mImages()
, mOptionfile("")
, mCurrentOptionObject(0)
, mOptions()
{
   // parse the command line, check for help message, check for config file, parse config file
   parseCommandLine(argc, argv);
}

void COptions::parseCommandLine(int argc, char* argv[])
{
   // parse the actual command line, use boost::program_options for that
   // Declare the supported options.
   boost::program_options::options_description cmdLineOptions("Options");
   static_cast<void>(cmdLineOptions.add_options()
       ("help,h", "show help message")
       ("config-file,c", boost::program_options::value<std::string>(&mOptionfile), "set configuration file")
       ("apply,a",boost::program_options::value<std::vector<std::string> >(&mImageNames)->composing(), "apply recolor to image(s)")
       )
   ;

   boost::program_options::variables_map optionset;
   
   try
   {
      boost::program_options::store(
         boost::program_options::parse_command_line(argc, argv, cmdLineOptions), optionset);
      boost::program_options::notify(optionset);
   }
   // catch any exception, if exception is detected, set options to invalid
   catch (const boost::program_options::error& e)
   {
      std::cout << "Error: " << e.what() << std::endl;
      mIsValid = false;
   }

   if (mIsValid)
   {
      // now check the base option set - help and config file
      if (optionset.count("help"))
      {
         mIsHelpRequested = true;
         std::cout << cmdLineOptions << std::endl;
      }
      else
      {
         if (optionset.count("config-file"))
         {
            // parse the options in the option file
            parseOptionfile();
         }
         else
         {
            std::cout << "Error: No configuration specified." << std::endl;
            mIsValid = false;
         }
         if (optionset.count("apply") && !checkImageFiles())
         {
            mIsValid = false;
         }
      }
   }
}

void COptions::parseOptionfile(void)
{
   std::ifstream optionfile(mOptionfile.c_str());
   if (optionfile.is_open())
   {
      std::string line;
      std::cout << "Parse optionfile"<<std::endl;
      UInt32 lines = 0;
      while (!optionfile.eof())
      {
         static_cast<void>(getline(optionfile,line));
         lines++;
      }
      static_cast<void>(optionfile.seekg(0));
      UInt32 currentLine = 0;
      UInt32 currentShow = 0;
      while (!optionfile.eof())
      {
         static_cast<void>(getline(optionfile,line));
         currentLine++;
         if (lines > 79)
         {
            if ((currentLine * 79 / lines) != currentShow)
            {
               currentShow = currentLine * 79 / lines;
               std::cout << ".";
            }
         }
         else
         {
            std::cout << ".";
         }
         if (!parseLine(line))
         {
            mIsValid = false;
            std::cout << std::endl << "Error: Parsing error at line " << currentLine-1 << std::endl;
            break;
         }
      }
      static_cast<void>(validateCurrentOption());
      std::cout << std::endl;
   }
   else
   {
      std::cout << "Error: File " << mOptionfile << " not found." << std::endl;
      mIsValid = false;
   }
}

void COptions::setCurrentOption(const std::string& option)
{
   if (option == "color")
   {
      mCurrentOptionObject = new CColorOption;
   }
   else
   if (option == "replace")
   {
      mCurrentOptionObject = new CReplaceOption;
   }
   else
   if (option == "merge" || option == "combine")
   {
      mCurrentOptionObject = new CMergeOption;
   }
   else
   if (option == "generic")
   {
      mCurrentOptionObject = new CGenericOption;
   }
}

bool COptions::parseLine(std::string& line)
{
   bool result = true;
   // check if it contains a section mark, i.e. [something]
   const boost::regex section("\\[(.*)\\]");
   const boost::regex comment("^ *#(.*)");
   boost::smatch match;
   if (line != "" && !boost::regex_search(line,comment))
   {
      if (boost::regex_match(line, match, section))
      {
         // the line contains some section description, match[1] contains the name
         // check if the currently active option object is valid
         if (validateCurrentOption())
         {
            setCurrentOption(match[1].str());
         }
         else
         {
            result = false;
         }
      }
      else
      {
         // just add the line to the current option object to be created
         mCurrentOptionObject->addLine(line);
      }
   }
   return result;
}

bool COptions::validateCurrentOption(void)
{
   bool valid = true;
   if (mCurrentOptionObject)
   {
      valid = mCurrentOptionObject->isValid(this);
      if (valid)
      {
         mCurrentOptionObject->finalize(this);
         mOptions.push_back(mCurrentOptionObject);
      }
      else
      {
         if (mCurrentOptionObject)
         {
            delete mCurrentOptionObject;
            mCurrentOptionObject = 0;
         }
      }
   }   
   return valid;
}

CColorOption* COptions::getColorOption(const std::string& name)
{
   for (UInt32 i = 0; i < mOptions.size(); i++)
   {
      CColorOption* const c = dynamic_cast<CColorOption*>(mOptions[i]);
      if (c && c->name() == name)
      {
         return c;
      }
   }
   return 0;
}

CReplaceOption* COptions::getReplaceOption(const std::string& name)
{
   for (UInt32 i = 0; i < mOptions.size(); i++)
   {
      CReplaceOption* const r = dynamic_cast<CReplaceOption*>(mOptions[i]);
      if (r && r->name() == name)
      {
         return r;
      }
   }
   return 0;
}

CGenericOption* COptions::getGenericOption(void)
{
   for (UInt32 i = 0; i < mOptions.size(); i++)
   {
      CGenericOption* const g = dynamic_cast<CGenericOption*>(mOptions[i]);
      if (g)
      {
         return g;
      }
   }
   return 0;
}

bool COptions::checkImageFiles(void)
{
   bool result = true;
   for (UInt32 i = 0; i < mImageNames.size(); i++)
   {
      CImage image(mImageNames[i]);
      if (!image.isValid())
      {
         result = false;
      }
      else
      {
         mImages.push_back(image);
      }
   }
   return result;
}
