// This file is part of TTDXRecolor.
// Copyright (c) 2010, Uwe Domaratius

#ifndef COPTIONS_HPP
#define COPTIONS_HPP

#include <string>
#include <vector>
#include "GTypes.hpp"
#include "CImage.hpp"

class COptionObject;
class CColorOption;
class CReplaceOption;
class CGenericOption;

//lint -esym(1712,COptions)
class COptions
{
public:
   /// constructor
   COptions(int argc, char* argv[]);

   /** check whether a help message is to be displayed
    *  @return true if help message is to be displayed
    */
   const bool isHelpRequested(void) const;
   
   /** check if given options are valid
    *  @return true if options are valid, false otherwise
    */
   const bool isValid(void) const;
   
   /** get the options at the given index
    *  @param index the index of the option to be checked
    *  @return the option at the given index
    */
   COptionObject* const options(UInt32 index) const;

   /** get a color option with the given name
    *  @param name the name of the color option
    *  @return pointer to the color option, 0 if no fitting option exists
    */
   CColorOption* getColorOption(const std::string& name);

   /** get a replace option with the given name
    *  @param name the name of the replace option
    *  @return pointer to the replace option, 0 if no fitting option exists
    */
   CReplaceOption* getReplaceOption(const std::string& name);

   /** get the generic option (there must be only one)
    *  @return pointer to generic option object, 0 if none exists
    */
   CGenericOption* getGenericOption(void);

   /** erase the given option
    *  @param object pointer to option that is to be erased
    */
   void erase(const COptionObject* const object);

   /** get the number of options that have been defined
    *  @return number of options that have been defined
    */
   const UInt32 numberOfOptions(void) const;

   /** get the number of images on which to apply recoloring
    *  @return number of images on which to apply recoloring
    */
   const UInt32 numberOfImages(void) const;

   /** get image on which to apply recoloring
    *  @param index of the image which to work on
    *  @return the image on which to work
    */
   const CImage getImage(const UInt32 index) const;

private:
   /// holds state of param -h/--help
   bool mIsHelpRequested;
   /// holds the validity state of the options
   bool mIsValid;
   /// the name(s) of image files onto which to apply recoloring (optional)
   std::vector<std::string> mImageNames;
   /// the images on which to work
   std::vector<CImage> mImages;
   /// the name of the optionfile, given by cmd line
   std::string mOptionfile;
   /// the current option that is parsed
   COptionObject* mCurrentOptionObject;
   /// all options that have been defined
   std::vector<COptionObject*> mOptions;

   /** parses the given commandline
    *  @param argc number of cmd line arguments
    *  @param argv list of cmd line arguments
    */
   void parseCommandLine(int argc, char* argv[]);

   /// parse the configuration file that has to be supplied
   void parseOptionfile();

   /** parse a single line
    *  @param line the line to be parsed
    */
   bool parseLine(std::string& line);

   /** validate the current option that has just been parsed
    *  @return true if option is valid, false otherwise
    */
   bool validateCurrentOption(void);

   /// set the current option object
   void setCurrentOption(const std::string& option);

   /// check given image files if they are valid
   bool checkImageFiles(void);
};

inline
const bool COptions::isHelpRequested(void) const
{
   return mIsHelpRequested;
}

inline
const bool COptions::isValid(void) const
{
   return mIsValid;
}

inline
COptionObject* const COptions::options(const UInt32 index) const
{
   return mOptions[index];
}

inline
const UInt32 COptions::numberOfOptions(void) const
{
   return mOptions.size();
}

inline
void COptions::erase(const COptionObject* const object)
{
   for (std::vector<COptionObject*>::iterator i = mOptions.begin();
        i != mOptions.end();
        ++i)
   {
      if (*i == object)
      {
         static_cast<void>(mOptions.erase(i));
         break;
      }
   }
}

inline
const UInt32 COptions::numberOfImages(void) const
{
   return mImages.size();
}

inline
const CImage COptions::getImage(const UInt32 index) const
{
   return mImages[index];
}

#endif
