// This file is part of TTDXRecolor.
// Copyright (c) 2010, Uwe Domaratius

#include "CRecolorApplication.hpp"
#include "CRecolorTable.hpp"
#include "CReplaceOption.hpp"
#include "CGenericOption.hpp"

#include <fstream>
#include <iostream>

CRecolorApplication::CRecolorApplication(const COptions& options)
: mOptions(options)
, mTables()
, mVerbose(false)
{
}

void CRecolorApplication::writeTableDescription(const UInt32 k, 
                                                const UInt32 number,
                                                std::ofstream& output, 
                                                const CReplaceOption* const r) const
{
   if (r->name() != "")
   {
      output << "// Replace: " << r->name();
      if ((r->fromSize()) < (r->toSize()))
      {
         output << "(" << k << ")";
      }
      output << std::endl;
   }
   output << "// from : ";
   for (UInt32 j = 0; j < r->fromNameSize(); j++)
   {
      output << r->fromName(j) << " ";
   }
   output << std::endl << "// to :   ";
   for (UInt32 j = 0; j < r->toNameSize() / number; j++)
   {
      output << r->toName((r->toNameSize() / number)*k+j) << " ";
   }
   if (mVerbose)
   {
      output << std::endl << "// replaces: ";
      for (UInt32 j = 0; j < r->fromSize(); j++)
      {
         output << r->from(j) << " ";
      }
      output << std::endl << "// with:     ";
      for (UInt32 j = 0; j < r->fromSize(); j++)
      {
         output << r->to(k*r->fromSize()+j) << " ";
      }
   }
}

void CRecolorApplication::writeTable(std::ofstream& output, const CReplaceOption* const r)
{
   const UInt32 number = r->toSize()/r->fromSize();
   std::cout << "Creating " << number << " table(s)." << std::endl;
   for (UInt32 k = 0; k < number; k++)
   {
      if (number > 100)
      {
         if (99 == k % 100)
         {
            std::cout << k+1 << "th table" << std::endl;
         }
      }
      CRecolorTable table;
      for (UInt32 j = 0; j < r->fromSize(); j++)
      {
         std::string index("0x");
         index += r->from(j);
         std::string value("0x");
         value += r->to(r->fromSize()*k+j);
         table.setEntry(NConvert::string2Int(index),NConvert::string2Int(value));
      }
      table.setPalette(r->isDos());
      mTables.push_back(table);
      writeTableDescription(k,number,output,r);
      output << std::endl << "-1 * 257 00 " << std::uppercase;
      // write recolor information
      for (UInt32 j = 0; j < 256; j++)
      {
         output << std::uppercase << std::hex << std::setfill('0') << std::setw(2) << table.getEntry(j) << " ";
      }
      output << std::endl << std::endl;
   }
}

int CRecolorApplication::run(void)
{
   const CGenericOption* const g = mOptions.getGenericOption();
   std::string outputName("recolor.txt");
   if (g)
   {
      outputName = g->output();
      mVerbose = g->verbose();
   }
   std::ofstream output(outputName.c_str());
   if (output.is_open())
   {
      // for each replace option, create and write one table
      for (UInt32 i = 0; i < mOptions.numberOfOptions(); i++)
      {
         const CReplaceOption* const r = dynamic_cast<CReplaceOption*>(mOptions.options(i));
         if (r)
         {
            writeTable(output,r);
         }
      }
      output.close();
      std::cout << "Tables written successfully." << std::endl;
   }
   else
   {
      std::cout << "Error: Cannot open output file " << outputName << "!" << std::endl;
   }
   applyOnImages();
   return 0;
}

void CRecolorApplication::applyOnImages()
{
   const UInt32 number = mOptions.numberOfImages();
   const UInt32 total = number*mTables.size();
   std::cout << "Creating " << total << " images." << std::endl;
   for (UInt32 i = 0; i < number; i++)
   {
      CImage image = mOptions.getImage(i);
      for (UInt32 j = 0; j < mTables.size(); j++)
      {
         if (total > 100)
         {
            const UInt32 current = i*mTables.size() + j;
            if (9 == current % 10)
            {
               std::cout << current+1 << "th image" << std::endl;
            }
         }
         image.applyRecolor(j,mTables[j]);
      }
   }
}
