// This file is part of TTDXRecolor.
// Copyright (c) 2010, Uwe Domaratius

#include "CReplaceOption.hpp"
#include "CColorOption.hpp"
#include "COptions.hpp"
#include "CGenericOption.hpp"
#include "CPalette.hpp"

#include <boost/regex.hpp>
#include <boost/tokenizer.hpp>
#include <iostream>

CReplaceOption::CReplaceOption()
: mName("")
, mFrom()
, mFromName()
, mTo()
, mToName()
, mDosPalette(false)
{
}

const UInt32 CReplaceOption::calculateIndexSize(const std::vector<std::string>& indices, COptions* options) const
{
   UInt32 number = 0;
   for (UInt32 i = 0; i < indices.size(); i++)
   {
      if (NConvert::string2Int(indices[i]) == INT_MAX)
      {
         const CColorOption* const c = options->getColorOption(indices[i]);
         if (c)
         {
            number += c->indices().size();
         }
      }
      else
      {
         number++;
      }
   }
   return number;
}

const bool CReplaceOption::colorsValid(const std::vector<std::string>& colors, COptions* options) const
{
   bool result = true;
   // check if given colors have been defined at all
   for (UInt32 i = 0; i < colors.size(); i++)
   {
      if ((NConvert::string2Int(colors[i]) == INT_MAX) && (!options->getColorOption(colors[i])))
      {
         std::cout << "Error: color " << colors[i] << " not found." << std::endl;
         result = false;
      }
   }
   return result;
}

bool CReplaceOption::isValid(COptions* options) const
{
   UInt32 error = 0;
   // at least one to and one from need to be given
   if (mTo.empty() || mFrom.empty())
   {
      error += E_REPLACE_UNDEFINED;
      printErrorMessage(E_REPLACE_UNDEFINED);
   }
   if (!colorsValid(mFrom,options) || !colorsValid(mTo,options))
   {
      error += E_REPLACE_COLOR_UNDEFINED;
   }

   // check if numbers of from and to fit together
   const UInt32 fromNumber = calculateIndexSize(mFrom,options);
   const UInt32 toNumber = calculateIndexSize(mTo,options);

   // sizes have to match
   if (fromNumber != 0 && toNumber % fromNumber != 0)
   {
      error += E_REPLACE_SIZE_MISMATCH;
      printErrorMessage(E_REPLACE_SIZE_MISMATCH);
      std::cout << "Number of To colors must be a multiple of number of From colors." << std::endl;
      std::cout << "Number of From colors: " 
         << mFrom.size() << ", number of To colors: " << mTo.size() << std::endl;
   }
   return 0 == error;
}

void CReplaceOption::finalize(COptions* options)
{
   // replace all color names by the appropriate indices
   // check if the result shall be dos or win -> if necessary indices need to be changed
   replaceColorNames(mFrom,mFromName, options);
   replaceColorNames(mTo,mToName, options);
}

void CReplaceOption::addLine(std::string& line)
{
   // a line always has to have the format option = value
   // non-matching lines are ignored
   const boost::regex option("([a-zA-Z]+)( ?)=( ?)(.+)");
   boost::smatch match;

   if (boost::regex_match(line, match, option))
   {
      std::string t = match[1].str();
      // I don't know why, but lint complains that it does not know std::transform
      static_cast<void>(std::transform(t.begin(), t.end(), t.begin(), tolower));//lint !e1025
      std::string t2 = match[4].str();
      static_cast<void>(std::transform(t2.begin(), t2.end(), t2.begin(), tolower));//lint !e1025
      if (t == "name")
      {
         mName = match[4].str();
      }
      if (t == "from" || t == "to")
      {
         // tokenize and push back to from-list
         // tokenize match[4] using whitespaces
         const std::string value = match[4].str();
         boost::char_separator<char> separator(" ");
         typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
         tokenizer tokens(value, separator);

         for (tokenizer::iterator i = tokens.begin();
              i != tokens.end(); ++i)
         {
            if (t == "from")
            {
               mFrom.push_back(*i);
            }
            else
            {
               mTo.push_back(*i);
            }
         }
      }
      if (t == "palette")
      {
         setPalette(match[4].str());
      }
   }
}


void CReplaceOption::setPalette(const std::string& palette)
{
   if (palette == "dos")
   {
      mDosPalette = true;
   }
   if (palette == "win")
   {
      mDosPalette = false;
   }
}

void CReplaceOption::mergeIndices(const UInt32 i, 
                                  const CReplaceOption* const option,
                                  std::vector<std::string>& intermediate) const
{
   if (option->toSize() / option->fromSize() > i)
   {
      const UInt32 base = i*option->fromSize();
      for (UInt32 j = 0; j < option->fromSize(); j++)
      {
         intermediate.push_back(option->to(base+j));
      }
   }
   else
   {
      for (UInt32 j = 0; j < option->fromSize(); j++)
      {
         intermediate.push_back(option->from(j));
      }
   }
}

void CReplaceOption::mergeNames(const UInt32 number, std::vector<std::string>& intermediate, const CReplaceOption* const rhs) const
{
   for (UInt32 i = 0; i < number; i++)
   {
      const UInt32 base = i*mFromName.size();
      for (UInt32 j = 0; j < mFromName.size(); j++)
      {
         intermediate.push_back(mToName[base+j]);
      }
      if (rhs->toSize() / rhs->fromSize() > i)
      {
         intermediate.push_back(rhs->toName(i));
      }
      else
      {
         intermediate.push_back(std::string(""));
      }
   }
}

void CReplaceOption::merge(const CReplaceOption* const rhs)
{
   if (rhs)
   {
      // for the to component, we need to do a bit more elaborate merging
      std::vector<std::string> intermediate;
      UInt32 numberFrom = mFrom.size();
      if (rhs->fromSize() > numberFrom)
      {
         numberFrom = rhs->fromSize();
      }
      // number of tables that will be generated is calculated
      // this also tells us how many sets we have to merge
      UInt32 numberTo = mTo.size() / mFrom.size();
      if (rhs->toSize() / rhs->fromSize() > numberTo)
      {
         numberTo = rhs->toSize() / rhs->fromSize();
      }

      // merge the indices
      for (UInt32 i = 0; i < numberTo; i++)
      {
         mergeIndices(i,this,intermediate);
         mergeIndices(i,rhs,intermediate);
      }
      mTo = intermediate;
      intermediate.clear();

      // merge the names
      mergeNames(numberTo,intermediate,rhs);
      mToName = intermediate;
      
      // the rest is simple concatenation
      for (UInt32 i = 0; i < rhs->fromSize(); i++)
      {
         mFrom.push_back(rhs->from(i));
      }
      for (UInt32 i = 0; i < rhs->fromNameSize(); i++)
      {
         mFromName.push_back(rhs->fromName(i));
      }
   }
}

void CReplaceOption::combineIndices(const UInt32 sets1,
                                    const UInt32 sets2, 
                                    const CReplaceOption* const rhs,
                                    std::vector<std::string>& intermediate) const
{
   for (UInt32 i = 0; i < sets1; i++)
   {
      const UInt32 base = i * mFrom.size();
      for (UInt32 j = 0; j < sets2; j++)
      {
         const UInt32 rhsbase = j*rhs->fromSize();
         for (UInt32 k = 0; k < mFrom.size(); k++)
         {
            intermediate.push_back(mTo[base+k]);
         }
         for (UInt32 k = 0; k < rhs->fromSize(); k++)
         {
            intermediate.push_back(rhs->to(rhsbase+k));
         }
      }
   }
}

void CReplaceOption::combineNames(const UInt32 sets1,
                                  const UInt32 sets2, 
                                  const CReplaceOption* const rhs,
                                  std::vector<std::string>& intermediate) const
{
   for (UInt32 i = 0; i < sets1; i++)
   {
      const UInt32 base = i*mFromName.size();
      for (UInt32 j = 0; j < sets2; j++)
      {
         const UInt32 rhsbase = (j*rhs->toNameSize() / sets2);
         for (UInt32 k = 0; k < mFromName.size(); k++)
         {
            intermediate.push_back(mToName[base+k]);
         }
         for (UInt32 k = 0; k < rhs->toNameSize() / sets2; k++)
         {
            intermediate.push_back(rhs->toName(rhsbase+k));
         }
      }
   }
}

void CReplaceOption::combine(const CReplaceOption* const rhs)
{
   if (rhs)
   {
      // for the to component, we need to do a bit more elaborate merging
      std::vector<std::string> intermediate;
      const UInt32 sets1 = mTo.size() / mFrom.size();
      const UInt32 sets2 = rhs->toSize() / rhs->fromSize();
      
      // combine indices
      combineIndices(sets1, sets2, rhs, intermediate);
      mTo = intermediate;
      intermediate.clear();

      // combine names
      combineNames(sets1, sets2, rhs, intermediate);
      mToName = intermediate;

      // the rest is simple concatenation
      for (UInt32 i = 0; i < rhs->fromSize(); i++)
      {
         mFrom.push_back(rhs->from(i));
      }
      for (UInt32 i = 0; i < rhs->fromNameSize(); i++)
      {
         mFromName.push_back(rhs->fromName(i));
      }
   }
}

void CReplaceOption::replaceColorNames(std::vector<std::string>& indices, std::vector<std::string>& names, COptions* options)
{
   std::vector<UInt32> intermediate;
   const CPalette targetPalette = CPalette(mDosPalette);

   for (std::vector<std::string>::iterator i = indices.begin(); i != indices.end(); ++i)
   {
      // for each entry, check if it is a defined color
      if (NConvert::string2Int(*i) == INT_MAX)
      {
         const CColorOption* const c = options->getColorOption(*i);
         if (c)
         {
            const CPalette fromPalette = CPalette(c->isDos());
            // if it is a defined color, get its indices
            for (UInt32 k = 0; k < c->indices().size(); k++)
            {
               // this palette conversion stuff neatly handles all palette conversions,
               // though most of the time it will surely convert from windows to windows...
               intermediate.push_back(targetPalette.getIndex(fromPalette.getColor(c->indices()[k])));
            }
            names.push_back(c->name());
         }
      }
      else
      {
         // it is not a color name, but a color index, so we'll just use it
         intermediate.push_back(NConvert::string2Int(*i));
         names.push_back(*i);
      }
   }
   indices.clear();
   for (UInt32 i = 0; i < intermediate.size(); i++)
   {
      indices.push_back(NConvert::int2String(intermediate[i]));
   }
   intermediate.clear();
}
