// This file is part of TTDXRecolor.
// Copyright (c) 2010, Uwe Domaratius

#ifndef CRECOLOROPTION_HPP
#define CRECOLOROPTION_HPP

#include <vector>
#include <string>

#include "GTypes.hpp"
#include "COptionObjects.hpp"

/// class describing a replace option from the config file
class CReplaceOption : public COptionObject
{
public:
   /// constructor
   CReplaceOption(void);

   /** returns if the option is valid
    *  @param options the list of previously encountered options
    *  @return true if option is valid, false otherwise
    */
   bool isValid(COptions* options) const;

   /** finalizes the option, setting up the index replacements in the correct way
    *  @param options list of previously encountered options
    */
   void finalize(COptions* options);

   /** add a line of the config file to the option
    *  @param line the line to be added
    */
   void addLine(std::string& line);

   /** get the colors that are to be replaced
    *  @param index the index to be looked up
    *  @return string indicating the color indices that are replaced
    */
   const std::string from(UInt32 index) const;

   /** get the number of colors given in the from vector
    *  @return number of colors given in the from vector
    */
   const UInt32 fromSize(void) const;

   /** get the colors that replace other colors
    *  @param index the index to be looked up
    *  @return string indicating the color index that replaces another color
    */
   const std::string to(UInt32 index) const;

   /** get the number of colors given in the to vector
    *  @return number of colors given in the to vector
    */
   const UInt32 toSize(void) const;

   /** get the name of the replacement
    *  @return the name of the replacement
    */
   const std::string name(void) const;

   /** get the name of the colors that are to be replaced
    *  @param index the index to be looked up
    *  @return string detailing the names of the colors that are replaced
    */
   const std::string fromName(UInt32 index) const;

   /** get the number of names given in the from name vector
    *  @return number of names given in the from name vector
    */
   const UInt32 fromNameSize(void) const;

   /** get the names of colors that replace other colors
    *  @param index the index to be looked up
    *  @return vector of strings detailing the names of colors that replace other colors
    */
   const std::string toName(UInt32 index) const;

   /** get the number of names given in the to name vector
    *  @return number of names given in the to name vector
    */
   const UInt32 toNameSize(void) const;

   /** merge the given replace option into this one
    *  @param rhs the option to be merged into this one
    */
   void merge(const CReplaceOption* const rhs);

   /** combine the given replace option with this one (construct each and every possible combination)
    *  @param rhs the option with which to do the combination
    */
   void combine(const CReplaceOption* const rhs);

   /** get the value of the dos flag
    *  @return flag indicating whether the output table is in dos palette
    */
   const bool isDos(void) const;
private:
   /// the name of the replace option
   std::string mName;
   /// the indices of the colors to be replaced
   std::vector<std::string> mFrom;
   /// the names of the colors to be replaced
   std::vector<std::string> mFromName;
   /// the indices of the colors that replace others
   std::vector<std::string> mTo;
   /// the names of the colors that replace others
   std::vector<std::string> mToName;
   /// indicates whether the output table should be in dos palette
   bool mDosPalette;

   /// replace the color names with the corresponding table indices
   void replaceColorNames(std::vector<std::string>& indices, std::vector<std::string>& name, COptions* options);

   /// calculate the number of color indices from the given vector which might also include color names
   const UInt32 calculateIndexSize(const std::vector<std::string>& indices, COptions* options) const;

   /// merge the indices of two replace options together
   void mergeIndices(const UInt32 i, 
                     const CReplaceOption* const option,
                     std::vector<std::string>& intermediate) const;

   /// merge the names of two replace options together
   void mergeNames(const UInt32 number, std::vector<std::string>& intermediate, const CReplaceOption* const rhs) const;

   /// check if given colors are valid
   const bool colorsValid(const std::vector<std::string>& colors, COptions* options) const;

   /// combine the color indices of two options
   void combineIndices(const UInt32 sets1,
                       const UInt32 sets2, 
                       const CReplaceOption* const rhs,
                       std::vector<std::string>& intermediate) const;

   /// combine the color names of two options
   void combineNames(const UInt32 sets1,
                     const UInt32 sets2, 
                     const CReplaceOption* const rhs,
                     std::vector<std::string>& intermediate) const;

   /// set the palette
   void setPalette(const std::string& palette);
};

inline
const std::string CReplaceOption::name(void) const
{
   return mName;
}

inline
const std::string CReplaceOption::from(UInt32 index) const
{
   return mFrom[index];
}

inline
const UInt32 CReplaceOption::fromSize(void) const
{
   return mFrom.size();
}

inline
const std::string CReplaceOption::to(UInt32 index) const
{
   return mTo[index];
}

inline
const UInt32 CReplaceOption::toSize(void) const
{
   return mTo.size();
}

inline
const std::string CReplaceOption::fromName(UInt32 index) const
{
   return mFromName[index];
}

inline
const UInt32 CReplaceOption::fromNameSize(void) const
{
   return mFromName.size();
}

inline
const std::string CReplaceOption::toName(UInt32 index) const
{
   return mToName[index];
}

inline
const UInt32 CReplaceOption::toNameSize(void) const
{
   return mToName.size();
}

inline
const bool CReplaceOption::isDos(void) const
{
   return mDosPalette;
}

#endif
